#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
TODO: Use https://python-saleae.readthedocs.io/en/latest/ to create captures which
can be presented to pytest as a fixture.
"""
