#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import sys

import nanaimo.cli

sys.exit(nanaimo.cli.main())
