#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
"""
The parser module contains parsers for common reporting formats used by
instruments or test frameworks.  These should be broken out from instruments
where appropriate to ensure reuse accross instrument families and communication
busses.
"""
